require 'forwardable'
require 'pg'

module HuggORM
  class Connection
    extend Forwardable

    RETRIES = 1

    attr_reader :options

    def self.db
      @_db ||= self.new
    end

    def connection
      return @connection unless @connection.nil?
      connect(@options)
      @connection
    end

    def options(opts = {})
      @options = opts
    end

    def connect(opts = {})
      @options = opts
      @connection = PG.connect(opts)
      self
    end

    def disconnect
      connection.finish if @connection
      @connection = nil
      self
    end

    def reconnect
      disconnect
      connect(@options)
      self
    end

    def method_missing(meth, *args, &block)
      if connection.respond_to?(meth)
        method_retry(meth, *args, &block)
      else
        super
      end
    end

    def method_retry(method, *args, &block)
      tries ||= RETRIES + 1
      connection.send(method, *args, &block)
    rescue PG::ConnectionBad, PG::AdminShutdown
      if (tries -= 1) > 0
        reconnect
        retry
      end
      raise
    end

    def exec(*args)
      method_retry(:exec, *args)
    end

    def exec_params(*args)
      method_retry(:exec_params, *args)
    end
  end
end
