module HuggORM
  class Query
    attr_reader :table, :column_name, :values,
                :_command, :_where, :_order, :_limit, :_offset, :_returning

    def initialize(table, column_name = 'data')
      @table = table
      @column_name = column_name
      @values = []
    end

    def run
      HuggORM::Connection.db.exec_params(to_sql, @values)
    end

    def to_sql
      sql = "#{@_command} #{@_where} #{@_order} #{@_limit} #{@_offset} #{@_returning}".strip
      sql.gsub!(/\s{2,}/, ' ')
      sql
    end

    def insert(data)
      @_command = "INSERT INTO #{@table} (#{@column_name}) VALUES (#{next_placeholder})"
      @values << data.to_json
      pg_result = run
      pg_result.cmd_tuples
    end

    def update(data, return_data = false)
      @_command = "UPDATE #{@table} SET #{@column_name} = #{@column_name} || #{next_placeholder}::jsonb"
      @values << data.to_json

      if return_data
        @_returning = "RETURNING #{@column_name}"
        return run.map { |tuple| deserialize(tuple["#{@column_name}"]) }
      end

      run.cmd_tuples
    end

    def select
      @_command = "SELECT #{@column_name} FROM #{@table}"
      pg_result = run
      pg_result.map { |tuple| deserialize(tuple["#{@column_name}"]) }
    end

    def delete
      @_command = "DELETE FROM #{@table}"
      run.cmd_tuples
    end

    def count
      @_command = "SELECT COUNT(*) FROM #{@table}"
      run.getvalue(0, 0).to_i
    end

    def where(conditions, values = [])
      values.size.times { |i| conditions.sub!(/\?/, next_placeholder(i + 1)) }
      @_where = "WHERE #{conditions}"
      @values += values
      self
    end

    def limit(limit, offset = nil)
      @_limit = "LIMIT #{next_placeholder}"
      @values << limit
      self.offset(offset) if offset
      self
    end

    def offset(offset)
      @_offset = "OFFSET #{next_placeholder}"
      @values << offset
      self
    end

    def order(*args)
      fields = []
      args.first.each_pair { |field, order| fields << "#{wrap_json(field)} #{order}" }
      @_order = "ORDER BY #{fields.join(', ')}"
      self
    end

    private

    def deserialize(data)
      JSON.parse(data, symbolize_names: true)
    end

    def next_placeholder(offset = 1)
      "$#{@values.size + offset}"
    end

    def wrap_json(field)
      "#{@column_name}->>'#{field}'"
    end
  end
end
